import sys
import os
from collections import defaultdict
from os import listdir
from os.path import (
    abspath,
    isdir,
    join,
    split,
)

import hammett.mark as mark


__version__ = '0.7.1'

from hammett.colors import (
    YELLOW,
    RED,
    RESET_COLOR,
    GREEN,
)

MISSING = object()

_orig_print = print


class Globals:
    def __init__(self):
        self.source_location = None
        self.modules = None
        self.orig_cwd = os.getcwd()
        self.verbose = False
        self.durations = False
        self.terminal_width = None
        self.results = None
        self.settings = {}
        self.fail_fast = False
        self.quiet = False
        self.drop_into_debugger = False
        self.durations_results = []
        self.disable_assert_analyze = False
        self.output = []
        self.result_db = None
        self.should_stop = False
        
    def reset(self):
        self.__init__()

    def get_log_without_colors(self):
        import re
        ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
        return ansi_escape.sub('', ''.join([
            str(args) + end
            for args, end, flush in self.output
        ]))


g = Globals()


def print(arg='', end='\n', flush=False):
    g.output.append((arg, end, flush))
    if g.quiet:
        return
    _orig_print(arg, end=end, flush=flush, file=sys.__stdout__)


def fixture(*args, **kwargs):
    args = list(args)

    def decorator(f):
        from hammett.impl import register_fixture
        register_fixture(f, *args, **kwargs)
        return f

    if len(args) == 1:
        if callable(args[0]):
            f = args[0]
            args[:] = args[1:]
            return decorator(f)
        else:
            return decorator
    else:
        assert not args
        return decorator


# This is a deprecated feature in pytest, but we'll keep it here for now. pytest-django uses it for example.
def yield_fixture(*args, **kwargs):
    return fixture(*args, **kwargs)


def raises(expected_exception, match=None):
    from hammett.impl import RaisesContext
    return RaisesContext(expected_exception, match=match)


def fail(message):
    raise RuntimeError(message)


class Option:
    def __init__(self):
        self.verbose = g.verbose


class Config:
    def __init__(self):
        self.option = Option()

    def getvalue(self, _):
        return None


class Request:
    current_fixture_setup = None

    def __init__(self, scope, parent, function=None):
        self.scope = scope
        self.function = function
        self.parent = parent
        self.additional_fixtures_wanted = set()
        self.keywords = {}
        self.fixturenames = set()
        self.finalizers = []
        self.fixture_results = {}
        self.funcargnames = []

        self.config = Config()

    def hammett_add_fixture_result(self, result):
        from hammett.impl import fixture_scope
        if fixture_scope.get(self.current_fixture_setup, 'function') != self.scope:
            self.parent.hammett_add_fixture_result(result)
        else:
            self.fixture_results[self.current_fixture_setup] = result

    def hammett_get_existing_result(self, name):
        try:
            return self.fixture_results[name]
        except KeyError:
            if self.parent:
                return self.parent.hammett_get_existing_result(name)
            return MISSING

    def teardown(self):
        for x in reversed(self.finalizers):
            x()

    @property
    def node(self):
        return self

    def addfinalizer(self, x):
        assert Request.current_fixture_setup is not None
        from hammett.impl import fixture_scope
        if fixture_scope.get(Request.current_fixture_setup, 'function') != self.scope:
            self.parent.addfinalizer(x)
        else:
            self.finalizers.append(x)

    def get_closest_marker(self, s):
        # TODO: fall back to the parent markers
        try:
            markers = self.function.hammett_markers
        except AttributeError:
            return None

        for marker in markers:
            if marker.name == s:
                return marker

        return None

    def getfixturevalue(self, s):
        return self.additional_fixtures_wanted.add(s)


def parse_markers(markers):
    if markers is None:
        return None

    def parse_one(s):
        if '[' not in s:
            return s, None

        assert s.endswith(']')
        base_name, _, name_plus_end_bracket = s.partition('[')
        return base_name, name_plus_end_bracket[:-1]

    # So far let's do something stupidly simple
    markers = markers.split(';')
    return dict(parse_one(m.strip()) for m in markers)


def guess_modules_and_source_path():
    this_dir = os.getcwd().split(os.sep)[-1]
    if isdir('lib'):
        return listdir('lib'), 'lib'
    elif isdir('src'):
        return listdir('src'), 'src'

    if isdir(this_dir):
        return [this_dir], '.'
    elif isdir(this_dir.replace('-', '_')):
        return [this_dir.replace('-', '_')], '.'
    elif isdir(this_dir.replace(' ', '_')):
        return [this_dir.replace(' ', '_')], '.'
    elif isdir(this_dir.replace('-', '')):
        return [this_dir.replace('-', '')], '.'
    elif isdir(this_dir.replace(' ', '')):
        return [this_dir.replace(' ', '')], '.'
    return [], '.'


def handle_dir(result, d):
    for root, dirs, files in os.walk(d):
        files = [
            f for f in files
            if f.endswith('__tests.py') or (f.startswith('test_') and f.endswith('.py'))
        ]
        result.extend(join(root, x) for x in files)


def collect_files(filenames):
    result = []
    if filenames is None:
        handle_dir(result, f'tests{os.sep}')
        handle_dir(result, f'test{os.sep}')
        for module_name in g.modules:
            handle_dir(result, join(g.source_location, module_name))
    else:
        for filename in filenames:
            if os.path.exists(filename):
                if os.path.isdir(filename):
                    handle_dir(result, filename)
                else:
                    result.append(filename)
            else:
                # This is a symbol, module or function/class? Try module first.
                symbol_path = join(g.source_location, filename.replace('.', os.sep))
                if os.path.exists(symbol_path + '.py'):
                    result.append(symbol_path + '__tests.py')

    return result


def collect_file_data(path):
    data = {}
    for root, dirs, files in os.walk(path):
        dirs[:] = [x for x in dirs if not x.startswith('.') and x not in ['venv', 'env', '__pycache__']]
        for filename in files:
            if not filename.endswith('.py'):
                continue
            full_path = join(root, filename)
            if full_path.startswith(f'.{os.sep}'):
                full_path = full_path[2:]
            data[full_path] = os.stat(full_path).st_mtime_ns
    return data


DB_VERSION = 1
DB_FILENAME = '.hammett-db'


def write_result_db(results):
    """
    The results database is a simple pickled dict with some keys:

    db_version: the version so we can change the format and throw away an old db if needed
    test_results: dict from filename -> (dict from test_name -> dict(stdout=str, stderr=str, status=str))
    file_data: dict filename -> nanosecond modification date
    """
    results['db_version'] = DB_VERSION
    from pickle import dump
    with open(DB_FILENAME, 'wb') as f:
        dump(results, f)


def new_result_db():
    return dict(
        db_version=DB_VERSION,
        test_results=defaultdict(dict),
        file_data=None,
    )


def read_result_db():
    from pickle import load
    try:
        with open(DB_FILENAME, 'rb') as f:
            results = load(f)
            if results['db_version'] != DB_VERSION:
                raise FileNotFoundError()
    except FileNotFoundError:
        return new_result_db()
    return results


def drop_cache_for_filename(result_db, filename):
    try:
        del result_db['test_results'][filename]
    except KeyError:
        pass


def update_result_db(result_db, new_file_data):
    if result_db['file_data'] is None:
        result_db['file_data'] = new_file_data
        assert not result_db['test_results']
        return result_db

    # Clear out test results when the test file or the tested module has changed
    old_file_data = result_db['file_data']
    clear_all_non_module_tests = False
    for filename, modification_time in old_file_data.items():
        if filename in new_file_data and modification_time != new_file_data[filename]:
            if filename.endswith('__tests.py') or filename.startswith('test_'):
                # The test has been changed
                drop_cache_for_filename(result_db, filename)
            else:
                # The module has been changed so translate the filename to the possible test files
                drop_cache_for_filename(result_db, filename[:-(len('.py'))] + '__tests.py')
                _, filename_only = split(filename)
                drop_cache_for_filename(result_db, f"tests{os.sep}{filename[:-(len('.py'))]}" + '__tests.py')
                clear_all_non_module_tests = True
                # TODO: this doesn't clear the db for module__function__tests.py

    if clear_all_non_module_tests:
        from pathlib import Path
        non_module_filenames = [x for x in result_db['test_results'].keys() if Path(x).stem.startswith('test_')]
        for x in non_module_filenames:
            del result_db['test_results'][x]

    result_db['file_data'] = new_file_data


def finish():
    for x in g.result_db['test_results'].values():
        for y in x.values():
            g.results[y['status']] += 1


def main(verbose=False, fail_fast=False, quiet=False, filenames=None, drop_into_debugger=False, match=None, durations=False, markers=None, disable_assert_analyze=False, module_unload=False, cwd=None):
    import sys
    if sys.version_info[:2] < (3, 6):
        print('hammett requires python 3.6 or later')
        exit(999)

    sys.modules['pytest'] = sys.modules['hammett']

    markers = parse_markers(markers)

    clean_up_sys_path = False

    if cwd is None:
        cwd = os.getcwd()
    cwd = abspath(cwd)
    g.orig_cwd = cwd
    os.chdir(cwd)

    if cwd not in sys.path:
        sys.path.insert(0, cwd)
        clean_up_sys_path = True

    from hammett.impl import should_stop
    g.reset()
    
    g.results = dict(success=0, failed=0, skipped=0, abort=0)
    g.verbose = verbose
    g.fail_fast = fail_fast
    g.quiet = quiet
    g.drop_into_debugger = drop_into_debugger
    g.durations = durations
    g.disable_assert_analyze = disable_assert_analyze

    from hammett.impl import read_settings, load_plugins
    read_settings()

    g.source_location = g.settings.get('source_location', '.')
    g.modules = g.settings.get('modules')
    if g.modules is None:
        m, s = guess_modules_and_source_path()
        g.source_location = g.source_location or s
        g.modules = m

    g.result_db = read_result_db()
    update_result_db(g.result_db, collect_file_data(g.source_location))
    write_result_db(g.result_db)

    filenames = collect_files(filenames)
    if not filenames:
        print('No tests found.')
        print('''You might need to add `modules` and `source_location` item under [hammett] in setup.cfg  like:

[hammett]
modules=
    foo
    bar
source_location=.
''')
        return 3

    plugins_loaded = False
    from os.path import split, sep

    session_request = Request(scope='session', parent=None)

    for test_filename in sorted(filenames):
        dirname, filename = split(test_filename)

        import importlib.util
        import sys
        if dirname.startswith(f'.{os.sep}'):
            dirname = dirname[2:]

        module_name = f'{dirname.replace(sep, ".")}.{filename.replace(".py", "")}'
        if module_name in sys.modules:
            del sys.modules[module_name]

        cache_filename = join(dirname, filename)
        if cache_filename in g.result_db['test_results']:
            continue

        # We do this here because if all test results are up to date, we want to avoid loading slow plugins!
        if not plugins_loaded:
            load_plugins()
            plugins_loaded = True

        spec = importlib.util.spec_from_file_location(module_name, test_filename)
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        try:
            spec.loader.exec_module(module)
        except Exception:
            print(f'Failed to load module {module_name}:')
            import traceback
            print(traceback.format_exc())
            g.results['abort'] += 1
            break

        module_request = Request(scope='module', parent=session_request)

        module_markers = getattr(module, 'pytestmark', [])
        if not isinstance(module_markers, list):
            module_markers = [module_markers]

        from hammett.impl import execute_test_function
        for name, f in list(module.__dict__.items()):
            if g.results['abort']:
                break
            if name.startswith('test_') and callable(f):
                for m in module_markers:
                    f = m(f)

                if match is not None:
                    if match not in name:
                        continue

                if markers is not None:
                    keep = False
                    for f_marker in getattr(f, 'hammett_markers', []):
                        if f_marker.name in markers:
                            arg = markers[f_marker.name]
                            if arg is not None:
                                assert len(f_marker.args) == 1, 'hammett only supports filtering on single arguments to markers right now'
                                if str(f_marker.args[0]) == arg:
                                    keep = True
                                    break
                            else:
                                keep = True
                                break
                    if not keep:
                        continue

                execute_test_function(module_name + '.' + name, f, module_request)
            if should_stop():
                break

        module_request.teardown()

        if module_unload:
            del sys.modules[module_name]

        if should_stop():
            break

    session_request.teardown()

    if not g.verbose:
        print()

    finish()

    color = GREEN
    if g.results['skipped']:
        color = YELLOW
    if g.results['failed']:
        color = RED

    if clean_up_sys_path:
        del sys.path[0]

    if g.durations:
        limit = 10
        print()
        print(f'--- {limit} slowest tests ---')
        g.durations_results.sort(key=lambda x: x[1], reverse=True)
        for name, time, _ in g.durations_results[:10]:
            print(f'{time}   {name}')

        print()
        print(f'--- {limit} slowest startup times for tests ---')
        print('Note: startup times might be shared cost so might show up unfairly')
        g.durations_results.sort(key=lambda x: x[2], reverse=True)
        for name, _, startup_time in g.durations_results[:10]:
            print(f'{startup_time}   {name}')

        print()

    print(f'{color}{g.results["success"]} succeeded, {g.results["failed"]} failed, {g.results["skipped"]} skipped{RESET_COLOR}')
    os.chdir(g.orig_cwd)

    if g.results['abort']:
        write_result_db(g.result_db)
        return 2

    write_result_db(g.result_db)
    return 1 if g.results['failed'] else 0


def hookimpl(*_, **__):
    print("WARNING: hookimpl is not implemented in hammett")
    return lambda f: f


def main_cli(args=None):
    if args is None:
        args = sys.argv[1:]
    from argparse import ArgumentParser
    parser = ArgumentParser(prog='hammett')
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true', default=False)
    parser.add_argument('-x', dest='fail_fast', action='store_true', default=False)
    parser.add_argument('-q', dest='quiet', action='store_true', default=False)
    parser.add_argument('-k', dest='match', default=None)
    parser.add_argument('-m', dest='markers', default=None)
    parser.add_argument('--durations', dest='durations', action='store_true', default=False)
    parser.add_argument('--no-assert-analyze', dest='disable_assert_analyze', action='store_true', default=False)
    parser.add_argument('--pdb', dest='drop_into_debugger', action='store_true', default=False)
    parser.add_argument(dest='filenames', nargs='*')
    args = parser.parse_args(args)

    return main(
        verbose=args.verbose,
        fail_fast=args.fail_fast,
        quiet=args.quiet,
        filenames=args.filenames or None,
        drop_into_debugger=args.drop_into_debugger,
        match=args.match,
        durations=args.durations,
        markers=args.markers,
        disable_assert_analyze=args.disable_assert_analyze,
    )


if __name__ == '__main__':
    exit(main_cli())



import os
import sys
from unittest import SkipTest
from datetime import datetime

import hammett
from hammett.colors import (
    GREEN,
    RED,
    YELLOW,
    MAGENTA,
    RESET_COLOR,
)


class RaisesContext(object):
    def __init__(self, expected_exception, match):
        self.expected_exception = expected_exception
        self.excinfo = None
        self.match = match

    def __enter__(self):
        self.excinfo = ExceptionInfo()
        return self.excinfo

    def __exit__(self, *tp):
        __tracebackhide__ = True
        if tp[0] is None:
            assert False, f'Did not raise {self.expected_exception}'
        if self.match:
            import re
            assert re.match(self.match, str(tp[1]))
        self.excinfo.value = tp[1]
        self.excinfo.type = tp[0]
        suppress_exception = issubclass(self.excinfo.type, self.expected_exception)
        return suppress_exception


class ExceptionInfo:
    def __init__(self):
        self.type = None
        self.value = None

    def __str__(self):
        return f'<ExceptionInfo: type={self.type} value={self.value}'


fixtures = {}
auto_use_fixtures = set()
fixture_scope = {}


def fixture_function_name(f):
    r = f.__name__
    if r == '<lambda>':
        import inspect
        r = inspect.getsource(f)
    return r


# TODO: store args
def register_fixture(fixture, *args, autouse=False, scope='function'):
    if scope == 'class':
        # hammett does not support class based tests
        return
    assert scope != 'package', 'Package scope is not supported at this time'

    name = fixture_function_name(fixture)
    # pytest uses shadowing.. I don't like it but I guess we have to follow that?
    # assert name not in fixtures, 'A fixture with this name is already registered'
    if hammett.g.verbose and name in fixtures and name != 'request':
        hammett.print(f'{fixture} shadows {fixtures[name]}')
    if autouse:
        auto_use_fixtures.add(name)
    assert scope in ('function', 'class', 'module', 'package', 'session')
    fixture_scope[name] = scope
    fixtures[name] = fixture


from hammett import fixtures as built_in_fixtures
for fixture_name, fixture in built_in_fixtures.__dict__.items():
    if not fixture_name.startswith('_'):
        register_fixture(fixture)


def pick_keys(kwargs, params):
    return {
        k: v
        for k, v in kwargs.items()
        if k in params
    }


class FixturesUnresolvableException(Exception):
    pass


_params_of_cache = {}


def params_of(f):
    if f not in _params_of_cache:
        import inspect
        _params_of_cache[f] = set(
            x.name
            for x in inspect.signature(f).parameters.values()
            if x.kind in (inspect.Parameter.POSITIONAL_OR_KEYWORD, inspect.Parameter.KEYWORD_ONLY)
        )
    return _params_of_cache[f]


def _teardown_yield_fixture(fixturefunc, it):
    """Executes the teardown of a fixture function by advancing the iterator after the
    yield and ensure the iteration ends (if not it means there is more than one yield in the function)"""
    try:
        next(it)
    except StopIteration:
        pass
    else:
        hammett.print(f"yield_fixture {fixturefunc} function has more than one 'yield'")
        exit(1)


def call_fixture_func(fixturefunc, request, kwargs):
    # TODO: this fixture crashes for some reason, blacklist it for now
    if fixturefunc.__name__ == '_dj_autoclear_mailbox':
        return None
    existing_result = request.hammett_get_existing_result(fixture_function_name(fixturefunc))
    if existing_result is not hammett.MISSING:
        return existing_result

    import inspect
    import functools

    from hammett import Request
    Request.current_fixture_setup = fixture_function_name(fixturefunc)

    res = fixturefunc(**kwargs)

    request.hammett_add_fixture_result(res)

    yieldctx = inspect.isgenerator(res)
    if yieldctx:
        it = fixturefunc(**kwargs)
        res = next(it)
        finalizer = functools.partial(_teardown_yield_fixture, fixturefunc, it)
        request.addfinalizer(finalizer)

    Request.current_fixture_setup = None

    return res


def dependency_injection(f, fixtures, orig_kwargs, request):
    fixtures = fixtures.copy()
    f_params = params_of(f)
    params_by_name = {
        name: params_of(fixture)
        for name, fixture in fixtures.items()
        # prune the fixture list based on what f needs
        if name in f_params or name in auto_use_fixtures or name == 'request'
    }

    def add_fixture(kwargs, name, indent=0):
        if name in kwargs:
            return

        params = params_of(fixtures[name])
        for param in params:
            add_fixture(kwargs, param, indent+1)
        params_by_name[name] = params

    kwargs = {}
    while params_by_name:
        reduced = False
        for name, params in list(params_by_name.items()):
            # If we have a dependency that we have pruned, unprune it
            for param in params:
                if param not in kwargs:
                    if param in fixtures:
                        params_by_name[param] = params_of(fixtures[param])
                        break
            # If we can resolve this dependency fully
            if params.issubset(set(kwargs.keys())):
                assert name not in kwargs
                kwargs[name] = call_fixture_func(fixtures[name], request, pick_keys(kwargs, params))
                if request is not None:
                    request.fixturenames.add(name)
                    for x in request.additional_fixtures_wanted:
                        add_fixture(kwargs, x)
                reduced = True
                del params_by_name[name]
        if not reduced:
            raise FixturesUnresolvableException(f'Could not resolve fixtures any more, have {params_by_name} left.\nAvailable dependencies: {kwargs.keys()}')

    if request is not None:
        request.fixturenames = set(kwargs.keys())
    # prune again, to get rid of auto use fixtures that f doesn't take
    kwargs = {k: v for k, v in kwargs.items() if k in f_params}
    return f, {**kwargs, **orig_kwargs}


def dependency_injection_and_execute(f, fixtures, orig_kwargs, request):
    f, kwargs = dependency_injection(f, fixtures, orig_kwargs, request)
    return f(**kwargs)


def should_stop():
    return hammett.g.should_stop


def should_skip(_f):

    try:
        for marker in _f.hammett_markers:
            if marker.name == 'skip':
                return True
    except AttributeError:
        pass

    return False


def indent(s, levels=1):
    ind = '    ' * levels
    return '\n'.join(ind + x for x in s.split('\n'))


def pretty_format(x, _indent=0):
    indent = '    ' * _indent
    result = ''
    if isinstance(x, dict):
        if not x:
            return '{}'
        result += '{\n'
        for k, v in x.items():
            result += f'{indent}    {k!r}: {pretty_format(v, _indent + 1)},\n'
        result += f'{indent}}}'
        return result
    elif isinstance(x, list):
        if not x:
            return '[]'
        result += '[\n'
        for v in x:
            result += f'{indent}    {pretty_format(v, _indent + 1)},\n'
        result += f'{indent}]'
        return result
    elif isinstance(x, tuple):
        if not x:
            return '(,)'
        result += '(\n'
        for v in x:
            result += f'{indent}    {pretty_format(v, _indent + 1)},\n'
        result += f'{indent})'
        return result
    else:
        return repr(x)


def feedback_for_exception():
    type, value, tb = sys.exc_info()
    while tb.tb_next:
        tb = tb.tb_next

    local_variables = tb.tb_frame.f_locals
    if local_variables:
        hammett.print('--- Local variables ---')
        for k, v in local_variables.items():
            hammett.print(f'{k}:')
            try:
                hammett.print(indent(pretty_format(v)))
            except Exception as e:
                hammett.print(f'   Error getting local variable repr: {e}')

    if type == AssertionError:
        analyze_assert(tb)


DIFF_STRING_SIZE_CUTOFF = 100


def analyze_assert(tb):
    if hammett.g.disable_assert_analyze:
        return

    # grab assert source line
    try:
        with open(tb.tb_frame.f_code.co_filename) as f:
            source = f.read().split('\n')
    except FileNotFoundError:
        try:
            with open(os.path.join(hammett.g.orig_cwd, tb.tb_frame.f_code.co_filename)) as f:
                source = f.read().split('\n')
        except FileNotFoundError:
            hammett.print(f'Failed to analyze assert statement: file not found. Most likely there was a change of current directory.')
            return

    line_no = tb.tb_frame.f_lineno - 1
    relevant_source = source[line_no]

    # if it spans multiple lines grab them all
    while line_no and not relevant_source.strip().startswith('assert '):
        line_no -= 1
        relevant_source = source[line_no] + '\n' + relevant_source

    if not relevant_source.strip().startswith('assert '):
        hammett.print('Failed to analyze assert statement (Did not find the assert)')
        return

    import ast
    try:
        assert_statement = ast.parse(relevant_source.strip()).body[0]
    except SyntaxError:
        try:
            # grab one more line after the the one where we got an exception, and try again
            relevant_source += '\n' + source[tb.tb_frame.f_lineno]
            assert_statement = ast.parse(relevant_source.strip()).body[0]
        except SyntaxError:
            hammett.print('Failed to analyze assert statement (SyntaxError)')
            return

    # We only analyze further if it's a comparison
    if assert_statement.test.__class__.__name__ != 'Compare':
        return

    hammett.print()
    hammett.print('--- Assert components ---')
    from astunparse import unparse
    try:
        left = eval(unparse(assert_statement.test.left), tb.tb_frame.f_globals, tb.tb_frame.f_locals)
        hammett.print('left:')
        hammett.print(indent(pretty_format(left)))
        right = eval(unparse(assert_statement.test.comparators), tb.tb_frame.f_globals, tb.tb_frame.f_locals)
    except Exception as e:
        hammett.print(f'Failed to analyze assert statement ({type(e)}: {e})')
        return
    hammett.print('right:')
    hammett.print(indent(pretty_format(right)))
    if isinstance(left, str) and isinstance(right, str) and len(left) > DIFF_STRING_SIZE_CUTOFF and len(right) > DIFF_STRING_SIZE_CUTOFF and '\n' in left:
        hammett.print()
        hammett.print('--- Diff of left and right assert components ---')
        left_lines = left.split('\n')
        right_lines = right.split('\n')
        from difflib import unified_diff
        for l in unified_diff(left_lines, right_lines, lineterm=''):
            color = ''
            if l:
                color = {
                    '+': GREEN,
                    '-': RED,
                    '@': MAGENTA,
                }.get(l[0], '')
            hammett.print(f'{color}{l}{RESET_COLOR}')


SKIPPED = 'skipped'
SUCCESS = 'success'
FAILED = 'failed'


MESSAGES = {
    SKIPPED: dict(
        v=f' {YELLOW}Skipped{RESET_COLOR}',
        s=f'{YELLOW}s{RESET_COLOR}',
    ),
    SUCCESS: dict(
        v=f' {GREEN}Success{RESET_COLOR}',
        s=f'{GREEN}.{RESET_COLOR}',
    ),
    FAILED: dict(
        v=f' {RED}Failed{RESET_COLOR}',
        s=f'{RED}F{RESET_COLOR}',
    ),
}
MSG_SKIPPED_VERBOSE = f' {YELLOW}Skipped{RESET_COLOR}'
MSG_SKIPPED = f'{YELLOW}s{RESET_COLOR}'
MSG_SUCCESS_VERBOSE = f' {GREEN}Success{RESET_COLOR}'
MSG_SUCCESS = f'{GREEN}.{RESET_COLOR}'


def inc_test_result(status, _name, _f, duration, stdout, stderr):
    verbose = hammett.g.verbose
    message = MESSAGES[status]['v' if verbose else 's']
    if verbose:
        hammett.print(message + (str(duration) if duration else ''))
    else:
        hammett.print(message, end='', flush=True)

    if hammett.g.fail_fast and status not in ('success', 'skipped'):
        hammett.g.should_stop = True

    filename = _f.__module__.replace('.', os.sep) + '.py'
    assert _name.startswith(_f.__module__)
    _name = _name[len(_f.__module__) + 1:]
    hammett.g.result_db['test_results'][filename][_name] = dict(stdout=stdout, stderr=stderr, status=status)


def run_test(_name, _f, _module_request, **kwargs):
    if should_skip(_f):
        inc_test_result(SKIPPED, _name, _f, duration=0, stdout='', stderr='')
        return

    from io import StringIO

    req = hammett.Request(scope='function', parent=_module_request, function=_f)

    def request():
        return req

    register_fixture(request, autouse=True)
    del request

    hijacked_stdout = StringIO()
    hijacked_stderr = StringIO()
    prev_stdout = sys.stdout
    prev_stderr = sys.stderr

    status = None
    duration = None
    start = None
    setup_time = None

    if hammett.g.verbose:
        hammett.print(_name + '...', end='', flush=True)
    try:
        sys.stdout = hijacked_stdout
        sys.stderr = hijacked_stderr

        if hammett.g.durations:
            start = datetime.now()

        resolved_function, resolved_kwargs = dependency_injection(_f, fixtures, kwargs, request=req)

        if hammett.g.durations:
            setup_time = datetime.now() - start
            start = datetime.now()

        resolved_function(**resolved_kwargs)

        if hammett.g.durations:
            duration = datetime.now() - start
            hammett.g.durations_results.append((_name, duration, setup_time))

        sys.stdout = prev_stdout
        sys.stderr = prev_stderr
        status = SUCCESS
    except KeyboardInterrupt:
        sys.stdout = prev_stdout
        sys.stderr = prev_stderr

        hammett.print()
        hammett.print('ABORTED')
        hammett.g.results['abort'] += 1
        hammett.g.should_stop = True
    except SkipTest:
        sys.stdout = prev_stdout
        sys.stderr = prev_stderr

        status = SKIPPED
    except:
        sys.stdout = prev_stdout
        sys.stderr = prev_stderr

        hammett.print(RED)
        if not hammett.g.verbose:
            hammett.print()
        hammett.print('Failed:', _name)
        hammett.print()

        import traceback
        hammett.print(traceback.format_exc())

        hammett.print()
        if hijacked_stdout.getvalue():
            hammett.print(YELLOW)
            hammett.print('--- stdout ---')
            hammett.print(hijacked_stdout.getvalue())

        if hijacked_stderr.getvalue():
            hammett.print(RED)
            hammett.print('--- stderr ---')
            hammett.print(hijacked_stderr.getvalue())

        hammett.print(RESET_COLOR)

        if not hammett.g.quiet:
            feedback_for_exception()

        if hammett.g.drop_into_debugger:
            try:
                import ipdb as pdb
            except ImportError:
                import pdb
            pdb.set_trace()

        status = FAILED

    assert status is not None
    inc_test_result(status, _name, _f, duration, stdout=hijacked_stdout.getvalue(), stderr=hijacked_stderr.getvalue())

    # Tests can change this which breaks everything. Reset!
    os.chdir(hammett.g.orig_cwd)
    req.teardown()


def execute_parametrize(_name, _f, _stack, _module_request, **kwargs):
    if not _stack:
        param_names = [f'{k}={v}' for k, v in kwargs.items()]
        _name = f'{_name}[{", ".join(param_names)}]'
        return run_test(_name, _f, _module_request=_module_request, **kwargs)

    names, param_list = _stack[0]
    if isinstance(names, str):
        names = [x.strip() for x in names.split(',')]
    for params in param_list:
        if not isinstance(params, (list, tuple)):
            params = [params]
        execute_parametrize(_name, _f, _stack[1:], _module_request, **{**dict(zip(names, params)), **kwargs})
        if should_stop():
            break


def execute_test_function(_name, _f, module_request):
    if getattr(_f, 'hammett_parametrize_stack', None):
        return execute_parametrize(_name, _f, _f.hammett_parametrize_stack, module_request)
    else:
        return run_test(_name, _f, module_request)


class FakePytestParser:
    def parse_known_args(self, *args):
        class Fake:
            pass
        s = Fake()
        s.itv = True
        s.ds = hammett.g.settings['django_settings_module']
        s.dc = None
        s.version = False
        s.help = False
        return s


class EarlyConfig:
    def __init__(self):
        self.inicfg = self
        self.config = self
        self.path = hammett.g.orig_cwd

    def addinivalue_line(self, name, doc):
        pass

    def getini(self, name):
        return None


early_config = EarlyConfig()


def load_plugin(module_name):
    import importlib
    try:
        plugin_module = importlib.import_module(module_name+'.plugin')
    except ImportError:
        plugin_module = importlib.import_module(module_name)

    parser = FakePytestParser()
    try:
        if hasattr(plugin_module, 'pytest_load_initial_conftests'):
            plugin_module.pytest_load_initial_conftests(early_config=early_config, parser=parser, args=[])

        if hasattr(plugin_module, 'pytest_configure'):
            import inspect
            if inspect.signature(plugin_module.pytest_configure).parameters:
                plugin_module.pytest_configure(early_config)
            else:
                plugin_module.pytest_configure()
    except Exception:
        hammett.print(f'Loading plugin {module_name} failed: ')
        import traceback
        hammett.print(traceback.format_exc())
        hammett.g.results['abort'] += 1
        hammett.g.should_stop = True
        return
    try:
        importlib.import_module(module_name + '.fixtures')
    except ImportError:
        pass
    return True


def read_settings():
    from configparser import (
        ConfigParser,
        NoSectionError,
    )
    config_parser = ConfigParser()
    config_parser.read('setup.cfg')
    try:
        hammett.g.settings.update(dict(config_parser.items('hammett')))
    except NoSectionError:
        return


def load_plugins():
    if 'plugins' not in hammett.g.settings:
        return

    import importlib
    for plugin in hammett.g.settings['plugins'].strip().split('\n'):
        load_plugin(plugin)
        if should_stop():
            return

    try:
        conftest = importlib.import_module('conftest')
    except ImportError:
        conftest = None

    if conftest is not None:
        plugins = getattr(conftest, 'pytest_plugins', [])
        for plugin in plugins:
            load_plugin(plugin)
            if should_stop():
                return



def parametrize(argnames, argvalues, indirect=False, ids=None, scope=None):
    def parametrize_wrapper(f):
        try:
            f.hammett_parametrize_stack.append((argnames, argvalues))
        except AttributeError:
            f.hammett_parametrize_stack = []
            f.hammett_parametrize_stack.append((argnames, argvalues))
        return f

    return parametrize_wrapper


class Marker:
    def __init__(self, name, args, kwargs):
        self.args = args
        self.kwargs = kwargs
        self.name = name

    def __repr__(self):
        return f'hammett.{self.name}'


def __getattr__(name: str):
    if name in __package__:
        return __package__[name]

    else:
        def marker(*args, **kwargs):
            args = list(args)

            def decorate(f):
                try:
                    f.hammett_markers.append(Marker(name, args, kwargs))
                except AttributeError:
                    f.hammett_markers = []
                    f.hammett_markers.append(Marker(name, args, kwargs))
                return f

            if len(args) == 1:
                if callable(args[0]):
                    f = args[0]
                    args[:] = args[1:]
                    return decorate(f)
            return decorate

        return marker
